import json
import logging
import pkgutil

from typing import Generic, TypeVar
from dataclasses import dataclass
from pydantic import BaseModel

from darq.app import Darq
from darq_ui.darq import Task, TaskStatus
from darq_ui.darq import DarqHelper
from darq_ui.utils import DarqUIConfig, join_url

log = logging.getLogger(__name__)

P = TypeVar("P", dict, BaseModel)


class Success(BaseModel, Generic[P]):
    payload: P | None = None
    error: None = None
    success: bool = True


class Failure(BaseModel, Generic[P]):
    payload: P | None = None
    error: str | None = None
    success: bool = False


def ok_response(
    payload: P | None = None,
) -> Success:
    return Success(
        error=None,
        payload=payload,
    )


def error_response(
    error: str | None = None,
) -> Failure:
    return Failure(
        error=error,
        payload=None,
    )


class TaskBody(BaseModel):
    name: str
    signature: str | None
    docstring: str | None
    status: TaskStatus | None
    dropped_reason: str | None

    class Config:
        use_enum_values = True


class TasksResponse(BaseModel):
    tasks: list[TaskBody]


class RunTaskResponse(BaseModel):
    task_id: str


class DropTaskResponse(BaseModel):
    task_name: str


class RemoveTaskFromDroplistResponse(BaseModel):
    task_name: str


@dataclass
class ErrorResult:
    error: str


@dataclass
class RunTaskResult:
    task_id: str


def get_index_page(ui_config: DarqUIConfig) -> str | None:
    try:
        page = pkgutil.get_data("darq_ui", "static/index.html")
    except FileNotFoundError:
        return None

    if not page:
        return None

    static_path = join_url(ui_config.base_path, "/static")
    config_str = json.dumps(ui_config.to_dict())

    page = page.replace(b"{{CONFIG}}", config_str.encode("utf-8"))
    page = page.replace(b"{{DYNAMIC_BASE}}", static_path.encode("utf-8"))

    return page.decode("utf-8")


async def get_tasks(darq_app: Darq) -> list[Task]:
    darq_helper = DarqHelper(darq_app)
    return await darq_helper.get_darq_tasks_for_admin()


async def run_task(
    darq_app: Darq,
    task_name: str,
    task_args: str | None = None,
    task_kwargs: str | None = None,
) -> ErrorResult | RunTaskResult:
    darq_helper = DarqHelper(darq_app)

    if await darq_helper.is_task_in_droplist(task_name):
        return ErrorResult(
            error="Task is in drop list and can not be run!",
        )

    try:
        t_args = json.loads(task_args or "[]")
    except json.JSONDecodeError:
        return ErrorResult(
            error="Failed to deserialize task args :(",
        )

    try:
        t_kwargs = json.loads(task_kwargs or "{}")
    except json.JSONDecodeError:
        return ErrorResult(
            error="Failed to deserialize task args :(",
        )

    darq_task = darq_helper.get_job_coro_by_name(task_name)

    if not darq_task:
        return ErrorResult(
            error=f'Task with name "{task_name}" does not exist!',
        )

    log.debug(
        "Run task %s with args %s and kwargs %s",
        task_name,
        t_args,
        t_kwargs,
    )

    job = await darq_task.delay(*t_args, **t_kwargs)

    if not job:
        return ErrorResult(
            error="Task failed to run",
        )

    return RunTaskResult(
        task_id=job.job_id,
    )


async def drop_task(darq_app: Darq, task_name: str, reason: str) -> None:
    """Stop running task by name and add it to a droplist.
    It can not be run again until removed from droplist."""
    darq_helper = DarqHelper(darq_app)
    await darq_helper.drop_add(task_name, reason)


async def remove_task_from_droplist(darq_app: Darq, task_name: str) -> None:
    darq_helper = DarqHelper(darq_app)
    await darq_helper.drop_remove(task_name)



from dataclasses import dataclass

DARQ_APP: str = "_darq_app"
DARQ_UI_CONFIG: str = "_darq_ui_config"


@dataclass
class DarqUIConfig:
    base_path: str
    logs_url: str | None
    embed: bool = False

    def to_dict(self) -> dict:
        return {
            "base_path": self.base_path,
            "logs_url": self.logs_url,
            "embed": self.embed,
        }


def join_url(base_url: str, path: str) -> str:
    """Join base url and path maintaining slashes."""
    return f"{base_url.rstrip('/')}/{path.lstrip('/')}"



import dataclasses
import pathlib
from typing import Any, Callable, Coroutine

from darq.app import Darq

from aiohttp import web
from aiohttp.web_response import json_response

from json import JSONDecodeError
from aiohttp.web_exceptions import HTTPNotAcceptable
from pydantic import BaseModel

from darq_ui.utils import DARQ_APP, DARQ_UI_CONFIG, DarqUIConfig, join_url
from darq_ui.handlers import (
    get_index_page,
    get_tasks,
    run_task,
    error_response,
    ok_response,
    Success,
    Failure,
    drop_task,
    remove_task_from_droplist,
    ErrorResult,
    TaskBody,
    TasksResponse,
    RunTaskResponse,
)


def get_darq_app(request: web.Request) -> Darq:
    return request.app[DARQ_APP]


async def get_json(request: web.Request) -> dict:
    if request.content_type == "application/json":
        try:
            return await request.json()
        except JSONDecodeError:
            pass
    raise HTTPNotAcceptable(text="invalid json")


def response_adapter(
    func: Callable,
) -> Callable[[web.Request], Coroutine[Any, Any, web.Response]]:
    """Allows to return Success or Failure objects from handlers"""

    async def wrapper(request: web.Request) -> web.Response:
        response = await func(request)
        if isinstance(response, Failure):
            return json_response(response.model_dump(), status=400)
        elif isinstance(response, BaseModel):
            return json_response(response.model_dump())
        return response

    return wrapper


async def index_handler(request: web.Request) -> web.Response:
    ui_config: DarqUIConfig = request.app[DARQ_UI_CONFIG]
    content = get_index_page(ui_config)
    if not content:
        return web.Response(text="not found", status=404)

    return web.Response(body=content, content_type="text/html")


async def embed_handler(request: web.Request) -> web.Response:
    ui_config: DarqUIConfig = request.app[DARQ_UI_CONFIG]
    content = get_index_page(dataclasses.replace(ui_config, embed=True))
    if not content:
        return web.Response(text="not found", status=404)

    return web.Response(body=content, content_type="text/html")


@response_adapter
async def get_tasks_handler(request: web.Request) -> TasksResponse:
    darq_app = get_darq_app(request)
    tasks = await get_tasks(darq_app)

    return TasksResponse(
        tasks=[
            TaskBody(
                name=task.name,
                signature=task.signature,
                docstring=task.doc,
                status=task.status,
                dropped_reason=task.dropped_reason,
            )
            for task in tasks
        ]
    )


@response_adapter
async def run_task_handler(
    request: web.Request,
) -> Success | Failure:
    darq_app = get_darq_app(request)
    data = await get_json(request)

    task_name = data.get("task_name")
    task_args = data.get("task_args")
    task_kwargs = data.get("task_kwargs")

    if not task_name:
        return error_response(error="task_name is required")

    if not task_args:
        return error_response(error="task_args is required")

    if not task_kwargs:
        return error_response(error="task_kwargs is required")

    result = await run_task(
        darq_app,
        task_name,
        task_args,
        task_kwargs,
    )

    if isinstance(result, ErrorResult):
        return error_response(
            error=result.error,
        )

    return ok_response(
        payload=RunTaskResponse(
            task_id=result.task_id,
        )
    )


@response_adapter
async def drop_task_handler(request: web.Request) -> Success | Failure:
    darq_app = get_darq_app(request)
    data = await get_json(request)

    task_name = data.get("task_name")
    reason = data.get("reason")

    if not task_name:
        return error_response(error="task_name is required")

    if not reason:
        return error_response(error="reason is required")

    await drop_task(
        darq_app,
        task_name,
        reason,
    )

    return ok_response()


@response_adapter
async def remove_task_from_droplist_handler(
    request: web.Request,
) -> Success | Failure:
    darq_app = get_darq_app(request)
    data = await get_json(request)

    task_name = data.get("task_name")

    if not task_name:
        return error_response(error="task_name is required")

    await remove_task_from_droplist(
        darq_app,
        task_name,
    )

    return ok_response()


def setup(
    app: web.Application,
    darq: Darq,
    base_path: str = "/darq",
    logs_url: str | None = None,
    web_ui: bool = True,
    embed: bool = False,
) -> None:
    """Setup Darq UI in aiohttp application.

    :param app: FastAPI application
    :param darq: Darq instance
    :param base_path: base path for Darq UI
    :param logs_url: URL to logs
    :param web_ui: enable web UI endpoint
    :param embed: enable /embed endpoint (for iframes)
    """
    if web_ui:
        app.router.add_get(base_path, index_handler)
    if embed:
        app.router.add_get(join_url(base_path, "/embed"), embed_handler)

    app.router.add_get(join_url(base_path, "/api/tasks"), get_tasks_handler)
    app.router.add_post(join_url(base_path, "/api/tasks/run"), run_task_handler)
    app.router.add_post(
        join_url(base_path, "/api/tasks/droplist/add"), drop_task_handler
    )
    app.router.add_post(
        join_url(base_path, "/api/tasks/droplist/remove"),
        remove_task_from_droplist_handler,
    )

    if web_ui or embed:
        here = pathlib.Path(__file__).parents[1]
        app.router.add_routes(
            [web.static(join_url(base_path, "/static"), here / "static")]
        )

    app[DARQ_APP] = darq
    app[DARQ_UI_CONFIG] = DarqUIConfig(base_path=base_path, logs_url=logs_url)



import pathlib
import dataclasses

from darq.app import Darq
from fastapi import FastAPI, Depends
from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles

from pydantic import BaseModel
from starlette.status import HTTP_404_NOT_FOUND

from darq_ui.utils import DARQ_APP, DARQ_UI_CONFIG, DarqUIConfig, join_url
from darq_ui.handlers import (
    get_tasks,
    run_task,
    drop_task,
    remove_task_from_droplist,
    get_index_page,
    error_response,
    ok_response,
    ErrorResult,
    TaskBody,
    TasksResponse,
    RunTaskResponse,
    DropTaskResponse,
    RemoveTaskFromDroplistResponse,
    Success,
    Failure,
)

from typing import Annotated


def get_darq_app(request: Request) -> Darq:
    return getattr(request.app, DARQ_APP)


def get_darq_ui_config(request: Request) -> DarqUIConfig:
    return getattr(request.app, DARQ_UI_CONFIG)


class RunTask(BaseModel):
    task_name: str
    task_args: str | None = None
    task_kwargs: str | None = None


class DropTask(BaseModel):
    task_name: str
    reason: str


class RemoveTaskFromDroplist(BaseModel):
    task_name: str


api_router = APIRouter()
index_router = APIRouter()


async def index_handler(
    ui_config: DarqUIConfig = Depends(get_darq_ui_config),
) -> HTMLResponse:
    content = get_index_page(ui_config)
    if not content:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No index.html found",
        )
    return HTMLResponse(content)


async def embed_handler(
    ui_config: DarqUIConfig = Depends(get_darq_ui_config),
) -> HTMLResponse:
    content = get_index_page(dataclasses.replace(ui_config, embed=True))
    if not content:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No index.html found",
        )
    return HTMLResponse(content)


@api_router.get("/tasks")
async def get_tasks_handler(
    darq_app: Annotated[Darq, Depends(get_darq_app)],
) -> TasksResponse:
    tasks = await get_tasks(darq_app)
    return TasksResponse(
        tasks=[
            TaskBody(
                name=task.name,
                signature=task.signature,
                docstring=task.doc,
                status=task.status,
                dropped_reason=task.dropped_reason,
            )
            for task in tasks
        ]
    )


@api_router.post(
    "/tasks/run",
    responses={
        200: {"model": Success[RunTaskResponse]},
        400: {"model": Failure},
    },
)
async def run_task_handler(
    darq_app: Annotated[Darq, Depends(get_darq_app)], task: RunTask
) -> Success | Failure:
    result = await run_task(
        darq_app,
        task.task_name,
        task.task_args,
        task.task_kwargs,
    )

    if isinstance(result, ErrorResult):
        return error_response(
            error=result.error,
        )

    return ok_response(
        payload=RunTaskResponse(
            task_id=result.task_id,
        )
    )


@api_router.post(
    "/tasks/droplist/add",
    responses={
        200: {"model": Success[DropTaskResponse]},
        400: {"model": Failure},
    },
)
async def drop_task_handler(
    darq_app: Annotated[Darq, Depends(get_darq_app)],
    task: DropTask,
) -> Success | Failure:
    """Stop running task by name and add it to a droplist.
    It can not be run again until removed from droplist."""
    await drop_task(
        darq_app,
        task.task_name,
        task.reason,
    )

    return ok_response()


@api_router.post(
    "/tasks/droplist/remove",
    responses={
        200: {"model": Success[RemoveTaskFromDroplistResponse]},
        400: {"model": Failure},
    },
)
async def remove_task_from_droplist_handler(
    darq_app: Annotated[Darq, Depends(get_darq_app)],
    task: RemoveTaskFromDroplist,
) -> Success | Failure:
    await remove_task_from_droplist(
        darq_app,
        task.task_name,
    )

    return ok_response()


def setup(
    app: FastAPI,
    darq: Darq,
    base_path: str = "/darq",
    logs_url: str | None = None,
    web_ui: bool = True,
    embed: bool = False,
) -> None:
    """Setup Darq UI in FastAPI application.

    :param app: FastAPI application
    :param darq: Darq instance
    :param base_path: base path for Darq UI.
        All api endpoints will be mounted under this path.
        Static files will be mounted under base_path + "/static".
    :param logs_url: URL to logs
    :param web_ui: enable web UI endpoint
    :param embed: enable /embed endpoint (for iframes)
    """

    # setup fastapi
    if web_ui:
        index_router.add_api_route(base_path, endpoint=index_handler)
    if embed:
        index_router.add_api_route(
            join_url(base_path, "/embed"), endpoint=embed_handler
        )

    if index_router.routes:
        app.include_router(index_router)

    app.include_router(api_router, prefix=join_url(base_path, "/api"))

    if web_ui or embed:
        here = pathlib.Path(__file__).parents[1]
        app.mount(
            join_url(base_path, "/static"),
            StaticFiles(
                directory=here / "static",
                html=True,
                check_dir=True,
            ),
            name="static",
        )

    setattr(app, DARQ_APP, darq)
    setattr(
        app,
        DARQ_UI_CONFIG,
        DarqUIConfig(base_path=base_path, logs_url=logs_url),
    )



import inspect
import logging
from datetime import datetime
from dataclasses import dataclass
from enum import Enum
from typing import cast

from darq.app import Darq
from darq.jobs import (
    DeserializationError,
    JobDef,
)
from darq.types import DarqTask
from darq.worker import Task as WorkerTask


log = logging.getLogger(__name__)


DROP_TASKS_KEY = "darq:drop_tasks"
PAUSE_TASKS_KEY = "darq:pause_tasks"


class DarqTaskDroppedError(Exception):
    pass


@dataclass
class TaskInfo:
    signature: str
    doc: str


class TaskStatus(Enum):
    RUNNING = "running"
    DROPPED = "dropped"


@dataclass
class Task:
    name: str
    status: TaskStatus | None
    enqueue_time: str | None
    signature: str | None
    doc: str | None
    dropped_reason: str | None = None


class DarqHelper:
    def __init__(self, darq_app: Darq) -> None:
        self.darq_app = darq_app

    async def get_drop_tasks(self) -> dict[str, str]:
        """Get all tasks that are in drop list.

        Returns a dictionary where key is task name and value is reason why task
        """
        assert self.darq_app.redis_pool
        raw_drop_tasks = await self.darq_app.redis_pool.hgetall(DROP_TASKS_KEY)
        return dict(raw_drop_tasks.items())

    def get_all_registered_jobs(self) -> list[str]:
        assert self.darq_app.redis_pool
        return sorted(self.darq_app.registry.keys())

    def get_darq_job_by_name(self, job_name: str) -> WorkerTask | None:
        assert self.darq_app.redis_pool
        return self.darq_app.registry.get(job_name)

    def get_job_coro_by_name(self, job_name: str) -> DarqTask | None:
        arq_function = self.get_darq_job_by_name(job_name)
        if not arq_function:
            return None
        return cast(DarqTask, arq_function.coroutine)

    def get_darq_tasks_info(self) -> dict[str, TaskInfo]:
        job_names = self.get_all_registered_jobs()

        task_info = {}
        for job_name in job_names:
            job = self.get_job_coro_by_name(job_name)
            if not job:
                continue
            signature = inspect.signature(job).replace(
                return_annotation=inspect.Signature.empty
            )

            # TODO(m.kind): remove ctx param from signature
            task_info[job_name] = TaskInfo(
                str(signature),
                job.__doc__ or "",
            )

        return task_info

    async def get_darq_tasks_for_admin(self) -> list[Task]:
        assert self.darq_app.redis_pool
        job_names = self.get_all_registered_jobs()

        queued_jobs: list = []

        try:
            queued_jobs = await self.darq_app.redis_pool.queued_jobs()
        except DeserializationError:
            log.exception("Can not get darq queued jobs")

        queued_jobs_map = {job.function: job for job in queued_jobs}

        tasks_info = self.get_darq_tasks_info()
        dropped_tasks = await self.get_drop_tasks()

        tasks = []
        for job_name in job_names:
            job_def: JobDef | None = queued_jobs_map.get(job_name)
            task_info = tasks_info[job_name]

            status = None
            if job_def:
                status = TaskStatus.RUNNING

            if job_name in dropped_tasks:
                status = TaskStatus.DROPPED

            task = Task(
                name=job_name,
                status=status,
                enqueue_time=job_def.enqueue_time.isoformat()
                if job_def
                else None,
                signature=task_info.signature,
                doc=task_info.doc,
                dropped_reason=dropped_tasks.get(job_name),
            )
            tasks.append(task)

        return tasks

    async def is_task_in_droplist(self, task_name: str) -> bool:
        assert self.darq_app.redis_pool
        return await self.darq_app.redis_pool.hexists(DROP_TASKS_KEY, task_name)

    async def drop_add(
        self, task_name: str, reason: str, user: str = "unknown"
    ) -> None:
        """Add task to drop list with reason and user who dropped task.

        Being in drop list means that task will be skipped on start.

        WARNING: it does not stop tasks which are already running. It only works
        on tasks that are not started yet or that are working in batches
        and running itselves.
        """
        assert self.darq_app.redis_pool
        add_date = datetime.now().strftime("%d.%m.%y %H:%M")
        reason = f"added to drop list by '{user}' on '{add_date}' because of '{reason}'"
        await self.darq_app.redis_pool.hset(DROP_TASKS_KEY, task_name, reason)

    async def drop_remove(self, task_name: str) -> None:
        """Remove task from drop list. After this operation a task can be
        started again."""
        assert self.darq_app.redis_pool
        await self.darq_app.redis_pool.hdel(DROP_TASKS_KEY, task_name)

    async def maybe_drop_task(self, task_name: str) -> None:
        """Check if task is in drop list and raise exception if it is.

        This function should be called before worker starts processing task.
        """
        assert self.darq_app.redis_pool
        if await self.is_task_in_droplist(task_name):
            reason = await self.darq_app.redis_pool.hget(
                DROP_TASKS_KEY, task_name
            )
            raise DarqTaskDroppedError(f"Task dropped: {reason}")


async def maybe_drop_task(darq: Darq, task_name: str) -> None:
    darq_helper = DarqHelper(darq)
    await darq_helper.maybe_drop_task(task_name)
